#!/usr/bin/python
# -*- coding: utf-8 -*-
import numpy as np

from auttitude.datamodels import LineSet
from auttitude.stats import DEFAULT_GRID


def michael(planes, lines):
    """Linear paleostress inversion by Michael (1984) method.

    Estimate paleostress from fault data using the method detailed in
    Michael (1984). Needs at least two faults (plane/line pairs) to work.

    Args:
        planes: (M, 3) array_like
                An array-like containing the direction cosines of the fault
                planes.
        lines: (M, 3) array_like
               An array-like containing the direction consines of the fault
               lines, respective to each plane.

    Returns:
        stress_matrix: (3, 3) ndarray
                       Least-squares solution to the inversion method.
        residual: float
                  Sum of residuals for the solution.

    Notes
    -----
    To linearize the inversion problem, _[1] assumes a priori that the
    tangential traction magnitude is constant on each fault plane, and as only
    relative stresses may be obtained by inversion, sets this traction to 1.
    This allows for a least squares solution to the problem.


    References
    ----------
    .. [1] Michael, A. J. (1984). Determination of stress from slip data:
           faults and folds. Journal of Geophysical Research: Solid Earth,
           89(B13), 11517-11526.
    """
    # pylint: disable=line-too-long, invalid-name
    A = np.zeros((3*len(planes), 5))
    for i, (n1, n2, n3) in enumerate(planes):
        A[3*i:3*(i+1), :] = ((n1 - n1**3 + n1*n3**2, n2 - 2*n2*n1**2, n3 - 2*n3*n1**2, -n1*n2**2 + n1*n3**2, -2*n1*n2*n3),
                             (-n2*n1**2 + n2*n3**2, n1 - 2*n1*n2**2, -2*n1*n2*n3, n2 - n2**3 + n2*n3**2, n3 - 2*n3*n2**2),
                             (-n3*n1**2 - n3 + n3**3, -2*n1*n2*n3, n1 - 2*n1*n3**2, -n2**2*n3 - n3 + n3**3, n2 - 2*n2*n3**2))
    stresses, residuals = np.linalg.lstsq(A, lines.ravel())[:2]
    s11, s12, s13, s22, s23 = -stresses
    stress_matrix = np.array(((s11, s12, s13),
                              (s12, s22, s23),
                              (s13, s23, -(s11 + s22))))
    return stress_matrix, residuals[0]


def angelier_graphical(planes, lines, grid=DEFAULT_GRID.grid):
    try:  # Try calculating directly with numpy arrays
        result = (grid.dot(np.transpose(planes))*grid.dot(np.transpose(lines))).sum(axis=1)*2
        return result
    except MemoryError:
        result = np.zeros((grid.shape[0], 1))
        for input_node, output_node in zip(grid, result):
            output_node[:] = 2*input_node.dot(planes)*input_node.dot(lines).sum()
    return result


def principal_stresses(stress_matrix):
    eigenvalues, eigenvectors = np.linalg.eigh(stress_matrix)
    eigenvalues_order = (eigenvalues).argsort()
    eigenvectors = LineSet(eigenvectors[:, eigenvalues_order].T)
    return eigenvectors, eigenvalues



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from math import acos, asin, atan2, cos, degrees, pi, radians, sin, sqrt

import numpy as np

from auttitude.io import (
    sphere_line,
    sphere_plane,
    translate_attitude,
    dcos_plane,
    dcos_line,
    format_attitude,
)
from auttitude.math import normalized_cross
from auttitude.stats import DEFAULT_GRID, SphericalStatistics


class Vector(np.ndarray):
    """
    Class that represents one normalized vector in space. This class 
    extends Numpy.ndarray class that is used as storage container for 
    the information.

    Parameters:
        dcos_data: Iterable object with 3 elements (can be even another 
        Vector) to construct a Vector from. Inside 'io' module there
        are methods to convert attitude data to normalized direction 
        cosines.
    """

    def __new__(cls, dcos_data):
        return np.asarray(dcos_data).view(cls)

    def angle_with(self, other, precise=False):
        """Returns the angle (in radians) between both vectors using
        the dot product between them. Optionally, if precise is set to True,
        calculates the angle using the arctangent of the ratio of the
        magnitude of the cross and dot products between the vectors.

        Parameter:
            other: A Vector like object.
            precise: whether to use arccosine or arctangent (defaults False)
        """
        if not precise:
            self_length = self.length
            other_length = sqrt(other.dot(other))
            return acos(self.dot(other) / (self_length * other_length))
        else:
            return atan2(self.cross_with(other), self.dot(other))

    def cross_with(self, other):
        """Returns the cross product between both vectors.

        Parameter:
            other: A Vector like object.
        """
        return Vector(np.cross(self, other))

    def normalized_cross_with(self, other):
        """Returns the normalized cross product between this and other
        vectors.

        Parameter:
            other: A Vector like object.
        """
        return Vector(normalized_cross(self, other))

    @property
    def attitude(self):
        """Returns the spherical coordinates of the normalized vector,
        considering it to be a Line in geological sense, as a
        Trend/Plunge pair in degrees."""
        x, y, z = self / self.length
        if z > 0:
            x, y = -x, -y
        return degrees(atan2(x, y)) % 360, degrees(asin(abs(z)))

    def format(
        self,
        strike=False,
        direction_format="quadrant",
        strike_format="right hand rule",
    ):
        """Returns the spherical coordinates of the normalized vector,
        considering it to be a Line in geological sense, as a
        Trend/Plunge pair formatted using the given parameters.
        Please refer to format_attitude method for description of parameters."""
        return format_attitude(
            *self.attitude,
            strike=strike,
            direction_format=direction_format,
            strike_format=strike_format
        )

    @property  # this should be cached
    def length(self):
        """Returns the euclidian norm of this vector."""
        return sqrt(self.dot(self))

    @property
    def direction_vector(self):
        """Returns the vector's left horizontal perpendicular vector.
        defaults to (1, 0, 0) if the vector is vertical."""
        if abs(self[2]) == 1.0:
            return Vector((1.0, 0.0, 0.0))
        direction = Vector((self[1], -self[0], 0.0))
        return direction / direction.length

    @property
    def dip_vector(self):
        """Returns the vector perpendicular to both this vector
        and it's direction vector. If this vector represents a plane,
        the resulting vector represents the maximum slope direction."""
        return Vector(np.cross(self / self.length, self.direction_vector))

    @property
    def projection_matrix(self):
        """Returns the matrix that projects vectors onto this vector."""
        return np.outer(self, self)

    @property
    def rejection_matrix(self):
        """Returns the matrix that rejects of a vector to this vector."""
        return np.eye(3) - self.projection_matrix

    @property
    def cross_product_matrix(self):
        """Returns the matrix that operates the cross product with this vector
        when multiplied by another vector"""
        return np.array(
            (
                (0.0, -self[2], self[1]),
                (self[2], 0.0, -self[0]),
                (-self[1], self[0], 0.0),
            )
        )

    def get_rotation_matrix(self, theta):
        """Returns the counterclockwise rotation matrix about this vector
        by angle theta.

        Parameters:
            theta: Rotation angle in radians
        """
        return (
            cos(theta) * np.eye(3)
            + sin(theta) * self.cross_product_matrix
            + (1 - cos(theta)) * self.projection_matrix
        )

    def get_great_circle(self, step=radians(1.0), offset=0.0):
        """Returns an array of n points equally spaced along the great circle
        normal to this vector.

        Parameters:
            step: Angular step in radians to generate points around great 
            circle.
            offset: Angular offset in radians from direction to generate points.
        """
        theta_range = np.arange(offset, 2 * pi + offset, step) % (2 * pi)
        sin_range = np.sin(theta_range)
        cos_range = np.cos(theta_range)
        return (
            (
                self.direction_vector[:, None] * cos_range
                + self.dip_vector[:, None] * sin_range
            ).T,
        )

    def get_small_circle(self, alpha, A=0, B=0, step=radians(1.0), offset=0.0):
        """Returns a pair of arrays representing points spaced step along
        both small circles with an semi-apical opening of alpha around
        this vector.

        Parameters:
            alpha: Apperture of the small circle in radians
            
            step: Angular step in radians to generate points around small
            circle.

            offset: Angular offset in radians from direction to generate points.
        """
        if A == 0 and B == 0:
            sc = self.get_great_circle(step, offset)[0].T * sin(alpha) + self[
                :, None
            ] * cos(alpha)
        else:
            theta_range = np.arange(0, 2 * pi, step)
            alpha_ = (
                alpha
                + A * np.cos(2 * theta_range)
                + B * np.sin(2 * theta_range)
            )
            sc = self.get_great_circle(step)[0].T * np.sin(alpha_) + self[
                :, None
            ] * np.cos(alpha_)
        return sc.T, -sc.T

    def arc_to(self, other, step=radians(1.0)):
        """Returns an array of points spaced step along the great circle
        between both vectors.

        Parameters:
            step: Angular step in radians to generate points along the 
            great-circle arc.
        """
        normal = self.rejection_matrix.dot(other)
        normal /= sqrt(normal.dot(normal))
        theta_range = np.arange(0, self.angle_with(other), step)
        sin_range = np.sin(theta_range)
        cos_range = np.cos(theta_range)
        return ((self * cos_range[:, None] + normal * sin_range[:, None]),)


class Plane(Vector):
    """
    Like the Vector class but, more specifically representing a plane in 
    space defined by the direction cosines of the plane dip direction/dip 
    pair.

    Parameters:
        dcos_data: Direction cosines of the dip direction/dip pair.
    """

    @staticmethod
    def from_attitude(direction, dip, strike=False):
        """
        Return a new Plane from direction, dip and strike given.
        Please refer to translate_attitude method for  parameters description.
        """
        dd, d = translate_attitude(direction, dip, strike)
        return Plane(dcos_plane((dd, d)))

    def intersection_with(self, other):
        """Returns the line of intersection of this and the other plane.

        Parameter:
            other: a Plane like object that will intersect with this object.
        """
        line = Line(self.cross_with(other))
        line_length = line.length
        return line / line_length if line_length > 0 else line

    @property
    def rhr_attitude(self):
        dd, d = self.attitude
        return (dd - 90) % 360, d

    @property
    def attitude(self):
        """Returns the spherical coordinates of the plane as a
        Dip Direction/Dip pair, in degrees."""
        x, y, z = self / self.length
        if z > 0:
            x, y = -x, -y
        return degrees(atan2(-x, -y)) % 360, degrees(acos(abs(z)))

    def format(
        self,
        strike=True,
        direction_format="azimuth",
        strike_format="dip quadrant",
    ):
        """Returns the spherical coordinates of the plane as a
        Direction/Dip pair formatted using the given parameters.
        Please refer to format_attitude method for description of parameters."""
        return format_attitude(
            *self.attitude,
            strike=strike,
            direction_format=direction_format,
            strike_format=strike_format
        )


class Line(Vector):
    """
    Like the Vector class but, more specifically representing a line in 
    space defined by the direction cosines of the line direction/dip.

    Parameters:
        dcos_data: Direction cosines of the line direction/dip.
    """

    @staticmethod
    def from_attitude(direction, dip, strike=False):
        """
        Return a new Line Object from direction, dip and strike given.
        Please refer to translate_attitude method for description of parameters.
        """
        direction, dip = translate_attitude(direction, dip, strike)
        return Line(dcos_line((direction, dip)))

    def plane_with(self, other):
        """Returns the plane containing this and the other line.

        Parameter:
            other: a Line like object that will define the returned plane.
        """
        plane = Plane(self.cross(other))
        plane_length = plane.length
        return plane / plane_length if plane_length > 0 else plane


class VectorSet(np.ndarray):
    """Class that represents a set (collection) of Vectors.

    Parameters:
        dcos_data: Is an array of direction cosines.
    """

    item_class = Vector

    def __new__(cls, dcos_data):
        obj = np.asarray(dcos_data).view(cls)
        return obj

    def __finalize_array__(self, obj):
        if obj is None:
            return

    def __getitem__(self, x):
        item = super(VectorSet, self).__getitem__(x)
        if np.atleast_2d(item).shape == (1, 3):
            return item.view(self.item_class)
        else:
            return item

    @property
    def stats(self):
        """Contains spherical statistics object for the data
        set.
        """
        return SphericalStatistics(self)

    @property
    def attitude(self):
        """Converts this data from direction cosines to attitudes."""
        return sphere_line(self)

    def count_fisher(self, k=None, grid=None):
        """Performs grid counting of the data by Fisher smoothing.

        Parameters:
            k: von Mises-Fisher k parameter, see 
            stats.SphericalGrid.count_fisher.
            
            grid: A stats.Spherical grid object to count on. If None
            the default grid defined on stats.DEFAULT_GRID will be
            used.
        """
        if grid is None:
            grid = DEFAULT_GRID
        return grid.count_fisher(self, k)

    def count_kamb(self, theta=None, grid=None):
        """Performs grid counting of the data by small circles of
        aperture theta.

        Parameters:
            theta: Robin and Jowett (1986) based on Kamb (1956) theta
            parameter, see stats.SphericalGrid.count_kamb.
            
            grid: A stats.Spherical grid object to count on. If None
            the default grid defined on stats.DEFAULT_GRID will be
            used.
        """
        if grid is None:
            grid = DEFAULT_GRID
        return grid.count_kamb(self, theta)

    def normalized_cross_with(self, other):
        """Returns a VectorSet object containing the normalized cross
        product of all possible pairs between this VectorSet and an
        (n, 3) array-like

        Parameter:
            other: A VectorSet like object.
        """
        vectors = np.zeros((len(self) * len(other), 3))
        i = 0
        for self_vector in self:
            for other_vector in other:
                cross = normalized_cross(self_vector, other_vector)
                vectors[i] = cross if cross[2] < 0 else -cross
                i += 1
        return VectorSet(vectors)

    def angle_with(self, other, precise=False):
        """Returns the angles matrix between this Spherical Data and an
        (n, 3) array-like.

        Parameter:
            other: A VectorSet like object.
            precise: whether to use arccosine or arctangent (defaults False)
        """
        angles = np.zeros((len(self), len(other)))
        for i, self_vector in enumerate(self):
            for j, other_vector in enumerate(other):
                angles[i, j] = self_vector.angle_with(other_vector, precise)
        return angles

    def get_great_circle(self, step=radians(1.0)):
        """Returns a generator to the list of great circles of 
        this VectorSet vectors.

        Parameters:
            step: Angular step in radians to generate points around great 
            circle.
        """
        for vector in self:
            yield vector.get_great_circle(step)[0]  # because of plot_circles


class PlaneSet(VectorSet):
    """Class that represents a set (collection) of Planes.

    Parameters:
        dcos_data: Is an array of direction cosines.
    """

    item_class = Plane

    def intersection_with(self, other):
        """Returns the intersection of all combinations of 
        planes in this set with the planes in other set as a
        list of lines defined as a VectorSet.

        Parameter:
            other: A PlaneSet like object.
        """
        return self.normalized_cross_with(other).view(LineSet)

    @property
    def attitude(self):
        """Converts this data from direction cosines to attitudes."""
        return sphere_plane(self)


class LineSet(VectorSet):
    """Class that represents a set (collection) of Lines.

    Parameters:
        dcos_data: Is an array of direction cosines.
    """

    item_class = Line

    def planes_with(self, other):
        """Return the list of Planes resulting from the 
        intersection of the combination of all lines in 
        this LineSet with other LineSet like object.

        Parameter:
            other: A LineSet like object.
        """
        return self.normalized_cross_with(other).view(PlaneSet)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from math import acos, atan2, cos, degrees, pi, sin, sqrt

import numpy as np

import auttitude as aut


def pi_transform(dcos_data):
    n = len(dcos_data)
    difference = np.zeros(int(n * (n - 1) / 2))
    for i in range(n):
        for j in range(i + 1, n):
            difference[i * n + j - 1] = dcos_data[i] - dcos_data[j]
    if issubclass(dcos_data, np.ndarray):
        return difference.view(dcos_data.__class__)
    else:
        return difference


def estimate_phi(data, lamb):
    """Phi estimation method by Fisher 1987."""
    a, b = 0.0, 0.0
    for point in data:
        a += sqrt(1.0 - np.dot(point, lamb)**2)
        b += abs(np.dot(point, lamb))
    return atan2(a, b)  # is this right?


def fisher_sc(data, lamb, maxiter=100, err=1e-6):
    """Automatic adjustment of small circle axis by the method of
    Fisher 1987 after Mardia 1977."""
    data_prime = np.zeros_like(data)
    for i in range(maxiter):
        phi = estimate_phi(data, lamb)
        for i, p in enumerate(data):
            data_prime[i] = (
                np.dot(p, lamb) * p - lamb) / sqrt(1 - np.dot(p, lamb)**2)
        Y = cos(phi) * np.sum(
            data_prime, axis=0) - sin(phi) * np.sum(
                data_prime, axis=0)
        lamb_ = Y / sqrt(np.dot(Y, Y))
        if acos(abs(np.dot(lamb, lamb_))) <= err:
            return lamb_, phi
        else:
            lamb = lamb_
    return lamb, phi


def stetsky_derivatives(dcos_data, parameter_array):
    trend, plunge, mu, A, B = parameter_array
    ct = cos(trend)
    st = sin(trend)
    cp = cos(plunge)
    sp = cos(plunge)
    ac = ct * sp
    bc = st * sp
    dc = cp
    derivatives = np.zeros((len(dcos_data), 5))
    for i, (ax, bx, dx) in enumerate(dcos_data):
        # ax, bx, dx = vector
        u = ac * ax + bc * bx + dc * dx
        squ = sqrt(1. - u**2)
        wn = -ax * st + bx * ct
        wd = ax * ct * cp + bx * st * cp - dx * sp
        w = wn / wd
        wl = atan2(wn, wd)
        dmdt = (ax * bc - bx * ac) / squ
        dmdp = -wd / squ
        wwd = wd * wd * (1. + w**2)
        dldt = (sp * (ax * dx * ct + bx * dx * st) - cp(ax**2 + bx**2)) / wwd
        dldp = -wn * u / wwd
        wl2 = 2 * wl
        xx = 2 * (A * sin(wl2) - B * cos(wl2))
        derivatives[i, 0] = dmdt + xx * dldt
        derivatives[i, 1] = dmdp + xx * dldp
        derivatives[i, 2] = -1.
        derivatives[i, 3] = -cos(wl2)
        derivatives[i, 4] = -sin(wl2)
    return derivatives


def stetsky_residual(dcos_data, parameter_array):
    trend, plunge, mu, A, B = parameter_array
    ct = cos(trend)
    st = sin(trend)
    cp = cos(plunge)
    sp = cos(plunge)
    ac = ct * sp
    bc = st * sp
    dc = cp
    residual = np.zeros(len(dcos_data))
    for i, (ax, bx, dx) in enumerate(dcos_data):
        amu = acos(ax*ac+bx*bc+dx*dc)
        wn = -ax * st + bx * ct
        wd = ax * ct * cp + bx * st * cp - dx * sp
        al2 = 2.*atan2(wn, wd)
        residual[i] = amu - A*cos(al2) - B*sin(al2)
    return residual

def stetsky_curfit(dcos_data, initial_guess, epsilon=0.01):
    parameter_array = np.radians(initial_guess)
    parameter_array[1] = pi/2. - parameter_array[1]
    n = len(dcos_data)
    dchi = 1000.
    while (dchi >= epsilon):
        lambda_ = 0.001
        beta = np.zeros(5)
        alpha = np.zeros((5,5))
        residual = stetsky_residual(dcos_data, parameter_array)
        derivatives = stetsky_derivatives(dcos_data, parameter_array)
        for i in range(len(n)):
            for j in range(5):
                beta[j] = beta[j] - residual[i]*derivatives[i, j]
                for k in range(j, 5):
                    alpha[j, k] = alpha[j, k] + derivatives[i, j]*derivatives[i, k]
                    alpha[k, j] = alpha[j, k]
        array = np.zeros((5,5))
        for j in range(5):
            for k in range(5):
                array[j, k] = alpha[j, k]/sqrt(alpha[j, j]*alpha[k, k])
            array[j, j] = 1. + lambda_
        arrayi = np.linalg.inv(array)
        
                




#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from auttitude.datamodels import Vector, Line, Plane, VectorSet, PlaneSet, LineSet
from auttitude.io import dcos_plane, sphere_plane, dcos_line, dcos_rake, sphere_line
from auttitude.stats import SphericalGrid, DEFAULT_GRID
from auttitude.io import translate_attitude

import auttitude.stats
import auttitude.math



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from math import asin, cos, degrees, log, pi, radians, sin, sinh, sqrt

import numpy as np

import auttitude as at
from auttitude.io import dcos_line, sphere_line


# maybe add option to build grid from projecting regular
# plane grid to sphere
class SphericalGrid(object):
    """
    This class represents an quasi-regular spherical grid with a given distance
    between nodes.

    Creates a hemi-spherical counting grid by tesselation.

        Parameters:
            node_spacing: Distance between nodes in degrees.

    """
    def __init__(self, node_spacing=2.5):
        nodes = [(0., 90.)]
        spacing = radians(node_spacing)
        for phi in np.arange(node_spacing, 90., node_spacing):
            azimuth_spacing = degrees(2 * asin(
                (sin(spacing / 2) / sin(radians(phi)))))
            for theta in np.linspace(0., 360. - azimuth_spacing, int(360. // azimuth_spacing)):
                nodes.append((theta + phi + node_spacing / 2, 90. - phi))
                nodes.append((theta - 180 + phi + node_spacing / 2, phi - 90.))
        for theta in np.arange(0., 360., node_spacing):
            nodes.append(((theta + 90. + node_spacing / 2) % 360., 0.))
        self.node_attitudes = nodes
        self.grid = dcos_line(np.array(nodes))

    @staticmethod
    def optimize_k(data):
        """Optimizes the value of K from the data, using Diggle and Fisher (88)
            method."""
        from scipy.optimize import minimize_scalar

        def obj(k):  # objective function to be minimized
            W = np.exp(k*(np.abs(np.dot(data, np.transpose(data)))))\
                * (k/(4*pi*sinh(k+1e-9)))
            np.fill_diagonal(W, 0.)
            return -np.log(W.sum(axis=0)).sum()

        return minimize_scalar(obj).x

    def count_fisher(self, data, k=None):
        """Performs axial data counting as in Robin and Jowett (1986).
        Will estimate an appropriate k if not given."""
        if k is None:
            k = self.optimize_k(data)  # Its a better estimate than R&J 86.
        return self.count(
            data,
            lambda nodes, data, k: np.exp(
                k * (np.abs(np.dot(nodes, np.transpose(data))) - 1)),
            k)

    def count_kamb(self, data, theta=None):
        """Performs data counting as in Robin and Jowett (1986) based on
        Kamb (1956). Will estimate an appropriate counting angle theta
        if not give."""
        if theta is None:
            theta = (len(data) - 1.0) / (len(data) + 1.0)
        else:
            theta = cos(radians(theta))
        return self.count(
            data,
            lambda nodes, data, theta: np.where(
                np.abs(np.dot(nodes, np.transpose(data))) >= theta, 1, 0),
            theta)

    def count(self, data, function, *args, **kwargs):
        """Generic counting grid method that accepts a function which
        receives the grid (or a node in the grid), the data and the additional
        arguments and keyword arguments passed to this method."""
        try:  # Try calculating directly with numpy arrays
            return function(self.grid, data, *args, **kwargs).sum(axis=1)
        except (MemoryError, ValueError):
            result = np.zeros(self.grid.shape[0])
            for i, input_node in enumerate(self.grid):
                result[i] = function(input_node, data, *args, **kwargs).sum()
        return result


class CircularGrid(object):
    def __init__(self, spacing=1., offset=0., **kwargs):
        self.spacing = spacing
        self.grid = self.build_grid(spacing, offset)

    def build_grid(self, spacing, offset=0., from_=0., to_=2 * pi):
        s = radians(spacing)
        o = radians(offset)
        theta_range = np.arange(o, 2 * pi + o, s)
        theta_range = theta_range[np.logical_and(theta_range >= from_,
                                                 theta_range <= to_)]
        return np.array((np.sin(theta_range), np.cos(theta_range))).T

    def cdis(self, data, nodes=None, axial=False):
        nodes = self.grid if nodes is None else nodes
        d = np.clip(
            np.dot(nodes, np.transpose(data)) / np.linalg.norm(data, axis=1),
            -1, 1)
        if axial:
            d = np.abs(d)
        return d

    def count(self,
              data,
              aperture=None,
              axial=False,
              spacing=None,
              offset=0,
              nodes=None,
              data_weight=None):
        aperture = radians(aperture) / 2. if aperture is not None else radians(
            self.spacing) / 2.
        if nodes is None:
            nodes = self.grid if spacing is None else self.build_grid(
                spacing, offset)
        spacing = radians(
            self.spacing) / 2 if spacing is None else radians(spacing) / 2
        c = cos(aperture)
        n = data.shape[0]
        data_weight = np.ones(n) if data_weight is None else data_weight
        return np.where(self.cdis(data, nodes, axial=axial) >= c,\
                        data_weight, 0.).sum(axis=1)[:,None]/data_weight.sum()

    def count_munro(self,
                    data,
                    weight=.9,
                    aperture=10.,
                    axial=False,
                    spacing=None,
                    offset=0,
                    nodes=None,
                    data_weight=None):
        spacing = 1 if spacing is None else spacing
        if nodes is None:
            nodes = self.grid if spacing is None else self.build_grid(
                spacing, offset)
        d = self.cdis(data, nodes, axial=axial)
        aperture = radians(aperture) / 2. if aperture is not None else radians(
            self.spacing) / 2.
        c = cos(aperture)
        theta = np.arccos(d) * pi / aperture
        data_weight = np.ones(
            data.shape[0]) if data_weight is None else data_weight
        upscale = 1. + 2. * np.power(
            weight, np.arange(0., aperture, radians(spacing))).sum()
        return (np.where(d >= c, data_weight, 0) * np.power(weight, theta)
                ).sum(axis=1)[:, None] * upscale / data_weight.sum()


class CircularStatistics(object):
    # pylint: disable=too-many-instance-attributes
    def __init__(self, data):  # Should this really be built by default?
        n = len(data)
        self.resultant_vector = at.datamodels.Vector(np.sum(data, axis=0))
        self.mean_resultant_vector = self.resultant_vector / n
        self.mean_vector = self.resultant_vector / self.resultant_vector.length
        self.resultant_length = self.resultant_vector.length
        self.mean_resultant_length = self.resultant_length / n

        self.resultant_vector_attitude = self.resultant_vector.attitude

        self.circular_variance = 1 - self.mean_resultant_length
        self.circular_standard_deviation = sqrt(
            -2 * log(1 - self.circular_variance))
        # self.circular_mean_direction_axial, self.circular_confidence_axial =\
        #     self.estimate_circular_confidence(axial=True)
        # self.circular_mean_direction, self.circular_confidence =\
        #     self.estimate_circular_confidence(axial=False)

        self.fisher_k = (n - 1) / (n - self.resultant_length)

        direction_tensor = np.dot(np.transpose(data), data) / n
        eigenvalues, eigenvectors = np.linalg.eigh(direction_tensor)
        eigenvalues_order = (-eigenvalues).argsort()

        self.eigenvalues = eigenvalues[eigenvalues_order]

        self.eigenvectors = [
            at.datamodels.Vector(eigenvector)
            for eigenvector in eigenvectors[:, eigenvalues_order].T
        ]
        self.eigenvectors_attitude = sphere_line(self.eigenvectors)


class SphericalStatistics(object):
    # pylint: disable=too-many-instance-attributes
    def __init__(self, data):  # Should this really be built by default?
        n = len(data)
        self.resultant_vector = at.datamodels.Vector(np.sum(data, axis=0))
        self.mean_resultant_vector = self.resultant_vector / n
        self.mean_vector = self.resultant_vector / self.resultant_vector.length
        self.resultant_length = self.resultant_vector.length
        self.mean_resultant_length = self.resultant_length / n

        self.resultant_vector_attitude = self.resultant_vector.attitude
        self.fisher_k = (n - 1) / (n - self.resultant_length)

        direction_tensor = np.dot(np.transpose(data), data) / n
        eigenvalues, eigenvectors = np.linalg.eigh(direction_tensor)
        eigenvalues_order = (-eigenvalues).argsort()

        lambda1, lambda2, lambda3 = self.eigenvalues =\
            eigenvalues[eigenvalues_order]
        lambda_sum = eigenvalues.sum()

        self.eigenvectors = [
            at.datamodels.Vector(eigenvector)
            for eigenvector in eigenvectors[:, eigenvalues_order].T
        ]
        self.eigenvectors_attitude = sphere_line(self.eigenvectors)

        # Check for divide by zero on stats?
        # From Vollmer 1990
        self.vollmer_P = (lambda1 - lambda2) / lambda_sum
        self.vollmer_G = 2 * (lambda2 - lambda3) / lambda_sum
        self.vollmer_R = 3 * lambda3 / lambda_sum

        self.vollmer_classification = ("point", "girdle", "random")[np.argmax((
            self.vollmer_P, self.vollmer_G, self.vollmer_R))]

        self.vollmer_B = self.vollmer_P + self.vollmer_G
        self.vollmer_C = log(lambda1 / lambda3)

        # From Woodcock 1977
        self.woodcock_Kx = log(lambda2 / lambda3)
        self.woodcock_Ky = log(lambda1 / lambda2)
        self.woodcock_C = log(lambda1 / lambda3)

        self.woodcock_K = self.woodcock_Ky / self.woodcock_Kx


def sample_fisher(mean_vector, kappa, n):
    """Samples n vectors from von Mises-Fisher distribution."""
    mean_vector = at.datamodels.Vector(mean_vector)
    direction_vector = mean_vector.direction_vector
    dip_vector = mean_vector.dip_vector
    kappa = kappa
    theta_sample = np.random.uniform(0, 2 * pi, n)
    alpha_sample = np.random.vonmises(0, kappa / 2., n)  # Why?
    return at.datamodels.VectorSet(
        ((direction_vector[:, None] * np.cos(theta_sample) +
          dip_vector[:, None] * np.sin(theta_sample)) * np.sin(alpha_sample) +
         mean_vector[:, None] * np.cos(alpha_sample)).T)


def sample_uniform(n):
    """Sample n vectors for the uniform distribution on the sphere."""
    samples = np.random.normal(size=(n, 3))
    return at.datamodels.VectorSet(
        samples / np.linalg.norm(samples, axis=1)[:, None])


DEFAULT_GRID = SphericalGrid(node_spacing=2.5)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
import re

import numpy as np

_DIRECTION_PATTERN = re.compile(r"([NS]?)([0-9.]*)([EW]?).*")
_DIP_PATTERN = re.compile(r"([0-9.]*)([NESW]*).*")


def process_dip(dip_value):
    """
    Parse dip values from string or float. Dip is defined as float number
    or string containing a number followed by two characters indicating the
    quadrant of dip that must match one of NE, SE, SW or NW.

    Quadrant of dip should be combined with with direction for proper
    interpretation of attitude. Please refer to the translate_attitude()
    method.

    Returns the dip value and quadrant.

    Parameters:
        dip_value: A string or float containing dip values to be parsed.
    """
    try:
        dip = float(dip_value)
        dip_quadrant = ""
    except ValueError:
        dip, dip_quadrant = _DIP_PATTERN.match(
            dip_value.strip().upper()
        ).groups()
        dip = float(dip)
        if dip_quadrant not in ("NE", "SE", "SW", "NW"):
            raise ValueError("invalid dip quadrant in: %s" % dip)
    return dip, dip_quadrant


def process_direction(direction_value):
    """
    Parse direction value from a Direction String or Float value. Floats are
    angles from North in clock-wise direction. Direction string encodes an 
    North or South (N or S), a float number and finally an optional East or West
    (E or W).
    
    Returns a direction value float number.
     
    Parameters:
        direction_value: a direction string or float value to be parsed.
    """
    try:
        direction = float(direction_value)
    except ValueError:
        leading, value, trailing = _DIRECTION_PATTERN.match(
            direction_value.strip().upper()
        ).groups()
        value = float(value)
        if leading == "N":
            if not trailing or trailing == "E":
                direction = value
            elif trailing == "W":
                direction = 360 - value
        elif leading == "S":
            if trailing == "E":
                direction = 180 - value
            elif trailing == "W":
                direction = 180 + value
            elif not trailing:
                raise ValueError("Invalid direction: %s" % direction)
        else:
            raise ValueError("Invalid direction: %s" % direction)
            # should these methods be more or less permissive>?
    return direction


def translate_attitude(direction, dip, strike=False):
    """
    Translate attitude values into proper direction and dip values using special
    methods for parsing spherical oriented notation for translation. Inside
    Auttitude dip is measured along direction unless stated (see strike option).
    
    Parameters:
        direction: A float or string value that can be interpreted as direction.
                   Please refer to io.parse_direction() method.

              dip: A float or string value that can be interpreted as dip.
                   Please refer to io.parse_dip() method.

           strike: If strike is true, this means that the direction GIVEN is not 
                   the direction of dipping, but, direction should be corrected
                   using the dipping quadrant or, if dipping quadrant is no given
                   should use the right-hand-rule meaning that attitude
                   direction (or direction of dipping) is 90 degrees further
                   (clock-wise) than indicated direction.
    """
    # Should the error messages be more verbose and explicitly say
    # why are they errors? Maybe yes.
    base_direction, base_dip = direction, dip
    dip, dip_quadrant = process_dip(dip)
    direction = process_direction(direction)
    if dip_quadrant:
        direction = direction % 180.0
        if 0 < direction < 90.0:
            if strike:
                if dip_quadrant == "SE":
                    direction = direction + 90.0
                elif dip_quadrant == "NW":
                    direction = direction + 270.0
                else:
                    raise ValueError(
                        "Invalid attitude: {}/{}".format(
                            base_direction, base_dip
                        )
                    )
            else:
                if dip_quadrant == "NE":
                    pass  # direction = direction
                elif dip_quadrant == "SW":
                    direction = direction + 180.0
                else:
                    raise ValueError(
                        "Invalid attitude: {}/{}".format(
                            base_direction, base_dip
                        )
                    )
        elif 90.0 < direction < 180.0:  # 90 <= direction < 180
            if strike:
                if dip_quadrant == "NE":
                    direction = direction - 90.0
                elif dip_quadrant == "SW":
                    direction = direction + 90.0
                else:
                    raise ValueError(
                        "Invalid attitude: {}/{}".format(
                            base_direction, base_dip
                        )
                    )
            else:
                if dip_quadrant == "SE":
                    pass  # direction = direction
                elif dip_quadrant == "NW":
                    direction = direction + 180.0
                else:
                    raise ValueError(
                        "Invalid attitude: {}/{}".format(
                            base_direction, base_dip
                        )
                    )
        elif direction == 0.0:
            if strike:
                if "E" in dip_quadrant:
                    direction = 90.0
                elif "W" in dip_quadrant:
                    direction = 270.0
                else:  # these else branches might be unreachable, check
                    raise ValueError(
                        "Invalid attitude: {}/{}".format(
                            base_direction, base_dip
                        )
                    )
            else:
                if "N" in dip_quadrant:
                    direction = 0.0
                elif "S" in dip_quadrant:
                    direction = 180.0
                else:
                    raise ValueError(
                        "Invalid attitude: {}/{}".format(
                            base_direction, base_dip
                        )
                    )
        elif direction == 90.0:
            if strike:
                if "N" in dip_quadrant:
                    direction = 0.0
                elif "S" in dip_quadrant:
                    direction = 180.0
                else:
                    raise ValueError(
                        "Invalid attitude: {}/{}".format(
                            base_direction, base_dip
                        )
                    )
            else:
                if "E" in dip_quadrant:
                    direction = 90.0
                elif "W" in dip_quadrant:
                    direction = 270.0
                else:
                    raise ValueError(
                        "Invalid attitude: {}/{}".format(
                            base_direction, base_dip
                        )
                    )
    elif strike:  # Right Hand Rule
        direction = (direction + 90.0) % 360.0

    return direction, dip


def dcos_plane(direction_dip):
    """
    Converts planes attitude (direction and dip measured from
    horizontal along direction) representing it as direction cosines of plane
    pole as used internally by Auttitude. Direction cosine values have norm
    equal to 1.
    
    Parameters:
        direction_dip: (2,N) iterable elements that contains values of plane
                       azimuth and dip measured from horizontal of N-planes to be
                       represented by its poles direction cosine. 
    """
    dd, d = np.transpose(np.radians(direction_dip))  # dip direction, dip
    return np.array(
        (-np.sin(d) * np.sin(dd), -np.sin(d) * np.cos(dd), -np.cos(d))
    ).T


def sphere_plane(dcos_data, rhr=False):
    """
    Converts to attitude of planes represented by its poles direction cosines.
    Attitudes on those cases are direction measured from North and Dip measured
    along direction unless rhr is set to True.
    
    Parameters:
        dcos_data: (3,N) direction cosines iterable element representing N-plane
                   poles direction cosines.

              rhr: Boolean indicating that direction should be corrected
                   considering the right-hand-rule.
    """
    x, y, z = np.transpose(dcos_data)
    sign_z = np.where(z > 0, 1, -1)
    z = np.clip(z, -1.0, 1.0)

    corr = 90.0 if rhr else 0.0
    return np.array(
        (
            ((np.degrees(np.arctan2(sign_z * x, sign_z * y)) - corr) % 360),
            np.degrees(np.arccos(np.abs(z))),
        )
    ).T


def dcos_line(trend_plunge):
    """
    Converts the attitude of lines (trend, plunge) into direction cosines values
    as used internally by Auttitude. Direction cosine values have norm equal to
    1.
    
    Parameters:
        trend_plunge: (2,N) elements iterable object that contains values for
                      line trend (orientation from North) and plunge (dipping
                      direction from horizontal) for N-lines to be represented
                      as directional cosines.
    """
    tr, pl = np.transpose(np.radians(trend_plunge))  # trend, plunge
    return np.array(
        (np.cos(pl) * np.sin(tr), np.cos(pl) * np.cos(tr), -np.sin(pl))
    ).T


def dcos_rake(direction_dip_rake):
    """
    Convert lines attitude (dip direction, dip and rake) into direction cosine
    values as used internally by Auttitude. Direction cosine values have norm
    equal to 1.
    
    Parameters:
        direction_dip_rake: (3,N) iterable elements that contains values for
                            direction, dip and rake. Dip is measured from 
                            Horizontal along Direction and Rake is measured
                            from direction.
    """
    dd, d, rk = np.transpose(np.radians(direction_dip_rake))  # trend, plunge
    return np.array(
        (
            np.sin(rk) * np.cos(d) * np.sin(dd) - np.cos(rk) * np.cos(dd),
            np.sin(rk) * np.cos(d) * np.cos(dd) + np.cos(rk) * np.sin(dd),
            -np.sin(rk) * np.sin(d),
        )
    ).T


def sphere_line(dcos_data):
    """
    Returns the attitude of lines from its direction cosines. Lines attitude 
    are defined as trend and plunge values.
    
    Parameters:
        dcos_data: (3,N) direction cosines iterable element representing N-line
                   direction cosines.
    """
    x, y, z = np.transpose(dcos_data)
    sign_z = np.where(z > 0, -1, 1)
    z = np.clip(z, -1.0, 1.0)
    return np.array(
        (
            np.degrees(np.arctan2(sign_z * x, sign_z * y)) % 360,
            np.degrees(np.arcsin(np.abs(z))),
        )
    ).T


def dcos_circular(direction, axial=False):
    """
    Convert circular attitudes (direction) into direction cosine
    values as used internally by Auttitude. Direction cosine values have norm
    equal to 1.
    
    Parameters:
        direction: (N)  iterable elements that contains values for
                        direction. Dip is measured from 
                        North clockwise.
        Axial: bool     Whether data is axial or vectorial. Defaults False.
    """
    if axial:
        direction = 2 * direction % 360.0
    d = np.radians(direction)
    return np.array((np.sin(d), np.cos(d), np.zeros_like(d))).T


def sphere_circular(dcos_data, axial=False):
    """
    Returns the attitude of circular data from its direction cosines. Circular
    attitude are defined as directions.
    
    Parameters:
        dcos_data: (3,N) direction cosines iterable element representing
                    N-circular direction cosines.
    """
    x, y = np.transpose(dcos_data)[:2]
    d = np.degrees(np.arctan2(x, y))
    if axial:
        direction = d / 2.0
    return direction


def direction_to_quadrant(d):
    d %= 360.0
    if 0.0 <= d < 90.0:
        return "N{}E".format(d)
    elif 90 <= d < 180.0:
        return "S{}E".format(180.0 - d)
    elif 180.0 <= d < 270.0:
        return "S{}W".format(d - 180.0)
    elif 270.0 <= d < 360.0:  # else:
        return "N{}W".format(360.0 - d)


def quadrant(d):
    d %= 360.0
    if 0.0 <= d < 90.0:
        return "NE"
    elif 90 <= d < 180.0:
        return "SE"
    elif 180.0 <= d < 270.0:
        return "SW"
    elif 270.0 <= d < 360.0:  # else:
        return "SW"


direction_formats = ["azimuth", "quadrant"]
strike_formats = ["right hand rule", "rhr", "dip quadrant"]


def format_attitude(
    direction,
    dip,
    strike=True,
    direction_format="quadrant",
    strike_format="dip quadrant",
    strike_input=False,
):
    direction_format = direction_format.lower()
    strike_format = strike_format.lower()
    direction, dip = translate_attitude(direction, dip, strike_input)
    if strike:
        if strike_format in ["right hand rule", "rhr"]:
            direction = (direction - 90.0) % 360.0
            dip_quadrant = ""
        elif strike_format in ["dip quadrant", "quadrant"]:
            dip_quadrant = quadrant(direction)
            direction = (direction - 90.0) % 180.0
        else:
            raise ValueError(
                "Strike format {} not recognized. Possible values are: {}".format(
                    strike_format, strike_formats
                )
            )
    else:
        dip_quadrant = ""
    if direction_format == "azimuth":
        direction_text = "{}".format(direction)
    elif direction_format == "quadrant":
        direction_text = direction_to_quadrant(direction)
    else:
        raise ValueError(
            "Direction format {} not recognized. Possible values are: {}".format(
                direction_format, direction_formats
            )
        )
    return direction_text, "{}{}".format(dip, dip_quadrant)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import

import abc

# http://treyhunner.com/2016/02/how-to-merge-dictionaries-in-python/
try:
    from collections import ChainMap
except ImportError:
    from itertools import chain

    def ChainMap(*args):
        return dict(chain(*map(lambda d: d.items(), reversed(args))))


from inspect import isclass
from math import cos, pi, radians, sin, sqrt

import numpy as np
from matplotlib import patheffects
from matplotlib.collections import LineCollection, PolyCollection
from matplotlib.lines import Line2D
from matplotlib.mlab import griddata
from matplotlib.patches import Circle, FancyArrowPatch

from auttitude.math import build_rotation_matrix
import auttitude as at


class ProjectionBase(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def _dtr(self, x, y, z):
        raise Exception()

    @abc.abstractmethod
    def _itr(self, X, Y):
        raise Exception()

    def __init__(self, rotation=None):
        # Maybe check if rotation is already a rotation matrix, and
        # sets it directly if so. Then a matrix generated by a
        # Vector could be used here.
        self.rotation = rotation
        if rotation is not None:
            self.R = build_rotation_matrix(*rotation)
            self.I = np.linalg.inv(self.R)
        else:
            self.R = self.I = np.eye(3)

    def _pre_direct(self, data, invert_positive, rotate):
        if rotate and self.rotation is not None:
            x, y, z = self.R.dot(np.transpose(data))
        else:
            x, y, z = np.transpose(data)
        d = 1.0 / np.sqrt(x * x + y * y + z * z)
        if invert_positive:
            c = np.where(z > 0, -1, 1) * d
            return c * x, c * y, c * z
        else:
            return d * x, d * y, d * z

    def _post_inverse(self, data, rotate):
        if rotate and self.rotation is not None:
            return np.transpose(self.I.dot(data))
        else:
            return np.transpose(data)

    def read_plot(self, X, Y):
        if X * X + Y * Y > 1.0:
            return ""
        vector = self.inverse((X, Y))
        vector = self.I.dot(vector)
        theta, phi = at.Vector(vector).attitude
        if phi >= 0.0:
            return "Pole: %05.1f/%04.1f\nLine: %05.1f/%04.1f" % (
                theta,
                phi,
                (theta - 180) % 360.0,
                90.0 - phi,
            )
        else:
            return ""

    def direct(self, data, invert_positive=True, rotate=True):
        return self._dtr(*self._pre_direct(data, invert_positive, rotate))

    def inverse(self, data, rotate=True):
        return self._post_inverse(self._itr(*np.transpose(data)), rotate)


class EqualAngle(ProjectionBase):
    def _dtr(self, x, y, z):
        """equal-angle (stereographic) projection.

        Projects a point from the unit sphere to a plane using
        stereographic projection"""
        return x / (1 - z), y / (1 - z)

    def _itr(self, X, Y):
        """inverse equal-angle (stereographic) projection.

        Inverts the projection of a point from the unit sphere
        to a plane using stereographic projection"""
        x = 2.0 * X / (1.0 + X * X + Y * Y)
        y = 2.0 * Y / (1.0 + X * X + Y * Y)
        z = (-1.0 + X * X + Y * Y) / (1.0 + X * X + Y * Y)
        return x, y, z


class Orthographic(ProjectionBase):
    def _dtr(self, x, y, z):
        """orthographic projection on z=0 plane."""
        return x, y

    def _itr(self, X, Y):
        """Inverse orthographic projection from z=0 plane to unit sphere."""
        x, y = X, Y
        z = np.sqrt(1 - x * x - y * y)
        return x, y, z


class EqualArea(ProjectionBase):
    def _dtr(self, x, y, z):
        """equal-area (schmidt-lambert) projection.

        Projects a point from the unit sphere to a plane using
        lambert equal-area projection, though shrinking the projected
        sphere radius to 1 from sqrt(2)."""
        return x * np.sqrt(1 / (1 - z)), y * np.sqrt(1 / (1 - z))

    def _itr(self, X, Y):
        """inverse equal-area (schmidt-lambert) projection.

        Inverts the projection of a point from the unit sphere
        to a plane using lambert equal-area projection, cosidering
        that the projected radius of the sphere was shrunk to 1 from
        sqrt(2)."""
        X, Y = X * sqrt(2), Y * sqrt(2)
        x = np.sqrt(1 - (X * X + Y * Y) / 4.0) * X
        y = np.sqrt(1 - (X * X + Y * Y) / 4.0) * Y
        z = -1.0 + (X * X + Y * Y) / 2
        return x, y, z


class ProjectionPlot(object):
    point_defaults = {"marker": "o", "c": "#000000", "ms": 3.0}

    line_defaults = {"linewidths": 0.8, "colors": "#4D4D4D", "linestyles": "-"}

    polygon_defaults = {
        "linewidths": 0.8,
        "edgecolors": "#4D4D4D",
        "facecolors": "#FF8000",
    }

    contour_defaults = {"cmap": "Reds", "linestyles": "-", "antialiased": True}

    arrow_defaults = {"lw": 1.0, "ls": "-"}

    net_gc_defaults = {
        "linewidths": 0.25,
        "colors": "#808080",
        "linestyles": "-",
    }

    net_sc_defaults = {
        "linewidths": 0.25,
        "colors": "#808080",
        "linestyles": "-",
    }

    text_defaults = {
        "family": "sans-serif",
        "size": "x-small",
        "horizontalalignment": "center",
    }

    @staticmethod
    def _clip_lines(data, z_tol=0.1):
        """segment point pairs between inside and outside of primitive, for
        avoiding spurious lines when plotting circles."""
        z = np.transpose(data)[2]
        inside = z < z_tol
        results = []
        current = []
        for i, is_inside in enumerate(inside):
            if is_inside:
                current.append(data[i])
            elif current:
                results.append(current)
                current = []
        if current:
            results.append(current)
        return results

    @staticmethod
    def _join_segments(segments, c_tol=radians(1.0)):
        """segment point pairs between inside and outside of primitive, for
        avoiding spurious lines when plotting circles."""
        all_joined = False
        while not all_joined and len(segments) > 1:
            all_joined = True
            segment = segments.pop(0)
            if abs(segment[-1].angle_with(segments[0][0])) < c_tol:
                segment.extend(segments.pop(0))
                all_joined = False
            elif abs(segment[0].angle_with(segments[0][-1])) < c_tol:
                segment_b = segments.pop(0)
                segment_b.extend(segment)
                segment = segment_b
                all_joined = False
            elif abs(segment[-1].angle_with(segments[0][-1])) < c_tol:
                segment.extend(reversed(segments.pop(0)))
                all_joined = False
            elif abs(segment[0].angle_with(segments[0][0])) < c_tol:
                segment_b = segments.pop(0)
                segment_b.extend(reversed(segment))
                segment = segment_b
                all_joined = False
            segments.append(segment)
        return segments

    @staticmethod
    def _close_polygon(projected_polygon):
        print(projected_polygon.shape)
        first = projected_polygon[0]
        last = projected_polygon[-1]
        mid = (first + last) / 2
        mid = mid / np.linalg.norm(mid)
        if np.dot(first, last) == 0.0:
            mid = np.array([first[1], -first[0]])
        if np.linalg.norm(first) > 1.0 and np.linalg.norm(last) > 1.0:
            return np.vstack(
                [projected_polygon, [2 * last, 3 * mid, 2 * first]]
            )
        return projected_polygon

    @staticmethod
    def _net_grid(gc_spacing=10.0, sc_spacing=10.0, n=360, clean_caps=True):
        theta = np.linspace(0.0, 2 * pi, n)
        gc_spacing, sc_spacing = radians(gc_spacing), radians(sc_spacing)
        if clean_caps:
            theta_gc = np.linspace(0.0 + sc_spacing, pi - sc_spacing, n)
        else:
            theta_gc = np.linspace(0.0, pi, n)
        gc_range = np.arange(0.0, pi + gc_spacing, gc_spacing)
        gc_range = np.hstack((gc_range, -gc_range))
        sc_range = np.arange(0.0, pi + sc_spacing, sc_spacing)
        i, j, k = np.eye(3)
        ik_circle = i[:, None] * np.sin(theta) + k[:, None] * np.cos(theta)
        great_circles = [
            (
                np.array((cos(alpha), 0.0, -sin(alpha)))[:, None]
                * np.sin(theta_gc)
                + j[:, None] * np.cos(theta_gc)
            ).T
            for alpha in gc_range
        ]
        small_circles = [
            (ik_circle * sin(alpha) + j[:, None] * cos(alpha)).T
            for alpha in sc_range
        ]
        if clean_caps:
            for cap in (0, pi):
                theta_gc = np.linspace(cap - sc_spacing, cap + sc_spacing, n)
                great_circles += [
                    (
                        np.array((cos(alpha), 0.0, -sin(alpha)))[:, None]
                        * np.sin(theta_gc)
                        + j[:, None] * np.cos(theta_gc)
                    ).T
                    for alpha in (0, pi / 2.0)
                ]
        return great_circles, small_circles

    def __init__(self, axis=None, projection=None, rotation=None, limits=None):
        if projection is None:
            self.projection = EqualArea(rotation)
        elif isclass(projection):
            self.projection = projection(rotation)
        else:
            self.projection = projection
        if axis is None:
            from matplotlib import pyplot as plt

            self.axis = plt.gca()
            self.clear_diagram(limits)
        else:
            self.axis = axis
            self.clear_diagram()

        self.axis.format_coord = self.projection.read_plot

    def clear_diagram(self, limits=None):
        """Clears the plot area and plot the primitive."""
        self.axis.cla()
        self.axis.axis("equal")
        if limits is not None:
            left, right, up, down = limits
        else:
            left, right, up, down = -1.1, 1.1, -1.1, 1.1
        self.axis.set_xlim(left, right)
        self.axis.set_ylim(up, down)
        self.axis.set_axis_off()
        self.plot_primitive()

    def plot_primitive(self):
        """Plots the primitive, center, NESW indicators and North if no
        rotation."""
        self.primitive = Circle(
            (0, 0),
            radius=1,
            edgecolor="black",
            fill=False,
            clip_box="None",
            label="_nolegend_",
        )
        self.axis.add_patch(self.primitive)
        # maybe add a dict for font options and such...
        if self.projection.rotation is None:
            self.axis.text(0.01, 1.025, "N", **self.text_defaults)
            x_cross = [0, 1, 0, -1, 0]
            y_cross = [0, 0, 1, 0, -1]
            self.axis.plot(
                x_cross, y_cross, "k+", markersize=8, label="_nolegend_"
            )

    def as_points(self, vectors, **kwargs):
        """Plot points on the diagram. Accepts and passes aditional key word
        arguments to axis.plot."""
        X, Y = self.projection.direct(vectors)
        # use the default values if not user input
        # https://stackoverflow.com/a/6354485/1457481
        options = ChainMap({}, kwargs, self.point_defaults)
        self.axis.plot(X, Y, linestyle="", **options)

    def as_lines(self, lines, **kwargs):
        """plot a list of lines"""
        # use the default values if not user input
        # https://stackoverflow.com/a/6354485/1457481
        options = ChainMap({}, kwargs, self.line_defaults)
        # should change this for better support of huge data
        projected_lines = [
            np.transpose(
                self.projection.direct(
                    segment, invert_positive=False, rotate=False
                )
            )
            for circle in lines
            for segment in self._join_segments(
                self._clip_lines(np.dot(circle, self.projection.R.T))
            )
        ]
        circle_collection = LineCollection(projected_lines, **options)
        circle_collection.set_clip_path(self.primitive)
        self.axis.add_collection(circle_collection)

    def as_polygons(self, polygons, **kwargs):
        """plot a list of polygons"""
        # use the default values if not user input
        # https://stackoverflow.com/a/6354485/1457481
        options = ChainMap({}, kwargs, self.polygon_defaults)
        # should change this for better support of huge data
        projected_polygons = [
            self._close_polygon(
                np.transpose(
                    self.projection.direct(
                        segment, invert_positive=False, rotate=False
                    )
                )
            )
            for circle in polygons
            for segment in self._join_segments(
                self._clip_lines(np.dot(circle, self.projection.R.T))
            )
        ]
        poly_collection = PolyCollection(projected_polygons, **options)
        poly_collection.set_clip_path(self.primitive)
        self.axis.add_collection(poly_collection)

    def as_contours(
        self,
        nodes,
        count,
        n_data,
        n_contours=10,
        minmax=True,
        percentage=True,
        contour_mode="fillover",
        resolution=250,
        **kwargs
    ):
        """Plot contours of a spherical count. Parameters are the counting
        nodes, the actual counts and the number of data points. Returns the
        matplotlib contour object for creating colorbar."""
        if percentage:
            count = 100.0 * count / n_data
        if minmax:
            intervals = np.linspace(count.min(), count.max(), n_contours)
        else:
            intervals = np.linspace(0, count.max(), n_contours)
        xi = yi = np.linspace(-1.1, 1.1, resolution)
        # maybe preselect nodes here on z tolerance
        # inside = nodes[:, -1] <= 0.1
        # nodes = nodes[inside]
        # count = count[inside]
        X, Y = self.projection.direct(nodes, invert_positive=False)
        # zi = griddata(X, Y, count, xi, yi, interp="linear")
        # use the default values if not user input
        # https://stackoverflow.com/a/6354485/1457481
        options = ChainMap({}, kwargs, self.contour_defaults)

        contour_fill, contour_lines = None, None
        if contour_mode in ("fillover", "fill"):
            contour_fill = self.axis.tricontourf(X, Y, count, levels=intervals, **options)
            for collection in contour_fill.collections:
                collection.set_clip_path(self.primitive)
        if contour_mode != "fill":
            contour_lines = self.axis.tricontour(X, Y, count, levels=intervals, **options)
            for collection in contour_lines.collections:
                collection.set_clip_path(self.primitive)

        return contour_fill if contour_fill is not None else contour_lines

    def text(self, vector, text, border=None, **kwargs):
        foreground = kwargs.pop("foreground", "w")
        options = ChainMap({}, kwargs, self.text_defaults)
        X, Y = self.projection.direct(vector)
        txt = self.axis.text(X, Y, text, **options)
        if border is not None:
            txt.set_path_effects(
                [
                    patheffects.withStroke(
                        linewidth=border, foreground=foreground
                    )
                ]
            )

    def base_net(
        self,
        gc_spacing=10.0,
        sc_spacing=10.0,
        n=360,
        gc_options=None,
        sc_options=None,
        clean_caps=True,
        plot_cardinal_points=True,
        cardinal_options=None,
    ):
        gc, sc = self._net_grid(gc_spacing, sc_spacing, n, clean_caps)
        gc_options = {} if gc_options is None else gc_options
        sc_options = {} if sc_options is None else sc_options
        cardinal_options = {} if cardinal_options is None else cardinal_options
        gc_options = ChainMap({}, gc_options, self.net_gc_defaults)
        self.as_lines(gc, **gc_options)
        sc_options = ChainMap({}, sc_options, self.net_sc_defaults)
        self.as_lines(sc, **sc_options)

        cardinal_options = ChainMap(
            {}, cardinal_options, {"verticalalignment": "center"}
        )
        if plot_cardinal_points and self.projection.rotation is not None:
            cpoints = np.array(
                (
                    (0.0, 1.0, 0.0),
                    (1.0, 0.0, 0.0),
                    (0.0, -1.0, 0.0),
                    (-1.0, 0.0, 0.0),
                )
            )
            c_rotated = np.dot(cpoints, self.projection.R.T)
            for i, (point, name) in enumerate(zip(c_rotated, "NESW")):
                if point[2] > 0:
                    continue
                self.text(
                    cpoints[i],
                    name,
                    border=2.0,
                    foreground="w",
                    **cardinal_options
                )

    def as_arrow_on_pole(
        self,
        planes,
        lines,
        sense=True,
        arrowsize=radians(10),
        arrowcolor="#4D4D4D",
        footwall=False,
        **kwargs
    ):
        options = ChainMap({}, kwargs, self.arrow_defaults)
        for plane, line in zip(planes, lines):
            arrow_from = (
                cos(arrowsize / 2.0) * plane + sin(arrowsize / 2.0) * line
            )
            arrow_to = (
                cos(-arrowsize / 2.0) * plane + sin(-arrowsize / 2.0) * line
            )
            if footwall:
                arrow_from, arrow_to = arrow_to, arrow_from
            X, Y = self.projection.direct((arrow_from, arrow_to))
            if not sense:
                self.axis.add_line(
                    Line2D(X, Y, c=arrowcolor, label="_nolegend_", **options)
                )
            else:
                a, b = (X[0], Y[0]), (X[1], Y[1])
                self.axis.add_patch(
                    FancyArrowPatch(
                        a,
                        b,
                        shrinkA=0.0,
                        shrinkB=0.0,
                        arrowstyle="->,head_length=2.5,head_width=1",
                        connectionstyle="arc3,rad=0.0",
                        mutation_scale=2.0,
                        ec=arrowcolor,
                        **options
                    )
                )

    def as_arrow_on_plane(
        self,
        planes,
        lines,
        sense=True,
        arrowsize=radians(10),
        arrowcolor="#4D4D4D",
        footwall=False,
        **kwargs
    ):
        self.as_arrow_on_pole(
            lines, planes, sense, arrowsize, arrowcolor, not footwall, **kwargs
        )


class CircularPlot(object):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

setup(
    name="auttitude",
    version="0.1.3",
    packages=find_packages(),

    install_requires=[
        'numpy',
        'matplotlib',
    ],

    # metadata for upload to PyPI
    author="Arthur Endlein",
    author_email="endarthur@gmail.com",
    description="library for analysis of structural geology data",
    license="MIT",
    keywords="geology attitude stereonet projection structural",
    url="https://github.com/endarthur/autti",
    dowload_url="https://github.com/endarthur/autti/archive/v0.1.3.tar.gz",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
)


#!/usr/bin/python
# -*- coding: utf-8 -*-
# pylint: disable=invalid-name
from __future__ import absolute_import
from math import acos, cos, pi, radians, sin, sqrt
import auttitude as at
import numpy as np


def normalized_cross(a, b):
    """
    Returns the normalized cross product between vectors.
    Uses numpy.cross().
    
    Parameters:
        a: First vector.
        b: Second vector.
    """
    c = np.cross(a, b)
    length = sqrt(c.dot(c))
    return c/length if length > 0 else c


def general_plane_intersection(n_a, da, n_b, db):
    """
    Returns a point and direction vector for the line of intersection
    of two planes in space, or None if planes are parallel.
    
    Parameters:
        n_a: Normal vector to plane A
         da: Point of plane A
        n_b: Normal vector to plane B
         db: Point of plane B
    """
    
    # https://en.wikipedia.org/wiki/Intersection_curve
    
    n_a = np.array(n_a)
    n_b = np.array(n_b)
    da  = np.array(da)
    db  = np.array(db)
    
    l_v = np.cross(n_a, n_b)
    norm_l = sqrt(np.dot(l_v, l_v))
    if norm_l == 0:
        return None
    else:
        l_v /= norm_l
    aa = np.dot(n_a, n_a)
    bb = np.dot(n_b, n_b)
    ab = np.dot(n_a, n_b)
    d_ = 1./(aa*bb - ab*ab)
    l_0 = (da*bb - db*ab)*d_*n_a + (db*aa - da*ab)*d_*n_b
    
    return l_v, l_0


def small_circle_intersection(axis_a, angle_a, axis_b, angle_b):
    """
    Finds the intersection between two small-circles returning zero, one or two 
    solutions as tuple.  
    
    Parameters:
         axis_a: Vector defining first circle axis
        angle_a: Small circle aperture angle (in radians) around axis_a 
         axis_b: Vector defining second circle axis
        angle_b: Small circle aperture angle (in radians) around axis_b
    """
    line = general_plane_intersection(axis_a, cos(angle_a),
                                      axis_b, cos(angle_b))
    if line is None:
        return ()
    l_v, l_0 = line
    # https://en.wikipedia.org/wiki/Line%E2%80%93sphere_intersection
    b = 2*l_v.dot(l_0)
    delta = b*b - 4*(l_0.dot(l_0) - 1)
    # Should the answers be normalized?
    if delta < 0:
        return ()
    elif delta == 0:
        return -b/2.,
    else:
        sqrt_delta = sqrt(delta)
        return l_0 + l_v*(-b - sqrt_delta)/2., l_0 + l_v*(-b + sqrt_delta)/2.


def build_rotation_matrix(azim, plng, rake):
    """
    Returns the rotation matrix that rotates the North vector to the line given 
    by Azimuth and Plunge and East and Up vectors are rotate clock-wise by Rake
    around the rotated North vector. 
    
    Parameters:
        azim: Line Azimuth from North (degrees).
        plng: Line Plunge measured from horizontal (degrees).
        rake: Rotation angle around rotated axis (degrees).
    """
    # pylint: disable=bad-whitespace
    azim, plng, rake = radians(azim), radians(plng), radians(rake)

    R1 = np.array((( cos(rake), 0.,        sin(rake)),
                   ( 0.,        1.,         0.      ),
                   (-sin(rake), 0.,        cos(rake))))

    R2 = np.array((( 1.,        0.,        0.       ),
                   ( 0.,        cos(plng), sin(plng)),
                   ( 0.,       -sin(plng), cos(plng))))

    R3 = np.array((( cos(azim), sin(azim), 0.       ),
                   (-sin(azim), cos(azim), 0.       ),
                   ( 0.,        0.,        1.       )))

    return R3.dot(R2).dot(R1)


def adjust_lines_to_planes(lines, planes):
    """
    Project each given line to it's respective plane. Returns the projected
    lines as a new LineSet and the angle (in radians) between each line and
    plane prior to projection.
    
    Parameters: 
        lines:  A LineSet like object with an array of n Lines
        planes: A PlaseSet like object with an array of n Planes 
    """
    
    lines  = at.LineSet(lines)
    planes = at.PlaneSet(planes)
    
    angles = np.zeros(len(lines))
    adjusted_lines = np.zeros_like(lines)
    for i, (line, plane) in enumerate(zip(lines, planes)):
        cos_theta = np.dot(line, plane)
        angles[i] = pi/2. - acos(cos_theta)
        adjusted_line = line - line*cos_theta
        adjusted_lines[i] = adjusted_line/sqrt(np.dot(adjusted_line,
                                                      adjusted_line))
    return adjusted_lines, angles
